import { __decorate, __metadata } from "tslib";
import { call, cancel, delay, fork } from 'redux-saga/effects';
import { reducer, saga, service } from '../core/annotations';
import { formatAsyncResponse, xhr } from '../core/xhr';
import { SagaEffect } from '../types/saga';
import { get, set } from '../utils/object';
import DefaultGlobalService from './GlobalService';
const now = () => Math.floor(Date.now() / 1000);
export const cacheKey = (key) => `oneki.cache.${key.replaceAll('.', '_')}`;
export const NO_EXPIRATION = Number.MAX_VALUE;
const expirationTime = (response, options) => {
    let ttl = undefined;
    if (options.ttl) {
        return now() + options.ttl;
    }
    const cacheControl = response.headers.get('cache-control');
    if (cacheControl) {
        const directives = cacheControl.split(',');
        for (const directive in directives) {
            const d = directive.trim();
            if (d === 'no-store' || d === 'no-cache') {
                return 0;
            }
            if (d.startsWith('max-age')) {
                const tokens = d.split('=');
                if (tokens.length === 2) {
                    ttl = parseInt(tokens[1]);
                }
            }
        }
    }
    if (ttl === undefined) {
        // todo, check in settings if default cache ttl
    }
    if (ttl === undefined) {
        return 0;
    }
    return now() + ttl;
};
let CacheEntryService = class CacheEntryService extends DefaultGlobalService {
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *delayLoading(key, delay_ms) {
        yield this.setLoading(key, delay_ms ? false : true, true);
        if (delay_ms) {
            yield delay(delay_ms);
            yield this.setLoading(key, true, true);
        }
    }
    /**
     * Save the object to cache in the redux store
     * Several metadata are atttached to the object (expireAt, loading, fetching)
     *
     * @param {object} cacheEntry : the object with metadata
     */
    setCache(key, cacheEntry) {
        set(this.state, cacheKey(key), cacheEntry);
    }
    setError(key, error) {
        set(this.state, `${cacheKey(key)}.error`, error);
        this.setLoading(key, false, false);
    }
    setLoading(key, loading, fetching) {
        set(this.state, `${cacheKey(key)}.loading`, loading);
        set(this.state, `${cacheKey(key)}.fetching`, fetching);
    }
    /**
     * Get the object from the server and save it in the cache (= redux store)
     *
     * @param {object} action :
     *    - onError: callback for any error
     *    - onSuccess: callback for any success
     * @param {object} context :
     *    - store: the redux store
     *    - router: the OnekiJS router of the application
     *    - settings: the full settings object passed to the application
     */
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *fetch(url, options = {}, force = false) {
        const { store } = this.context;
        let loadingTask = null;
        try {
            const cacheEntry = get(store.getState(), cacheKey(url));
            if (cacheEntry && !force) {
                if (cacheEntry.expireAt && cacheEntry.expireAt > now())
                    return;
                if (cacheEntry.fetching)
                    return;
            }
            loadingTask = yield fork([this, this.delayLoading], url, options.delayLoading);
            const fetcher = options.fetcher || xhr;
            const response = yield fetcher(url, 'GET', undefined, options);
            const result = yield formatAsyncResponse(response);
            yield cancel(loadingTask);
            yield this.setCache(url, {
                payload: result,
                expireAt: expirationTime(response, options),
                loading: false,
                fetching: false,
                error: undefined,
            });
            const onSuccess = options.onFetchSuccess;
            if (onSuccess) {
                yield call(onSuccess, result);
            }
        }
        catch (e) {
            if (process.env.NODE_ENV === 'development') {
                console.error('Fetch error', e);
            }
            if (loadingTask) {
                yield cancel(loadingTask);
            }
            yield this.setError(url, e);
            const onError = options.onFetchError;
            if (onError) {
                yield call(onError, e);
            }
            else {
                throw e;
            }
        }
    }
};
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", void 0)
], CacheEntryService.prototype, "setCache", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", void 0)
], CacheEntryService.prototype, "setError", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Boolean, Boolean]),
    __metadata("design:returntype", void 0)
], CacheEntryService.prototype, "setLoading", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object, Object]),
    __metadata("design:returntype", void 0)
], CacheEntryService.prototype, "fetch", null);
CacheEntryService = __decorate([
    service
], CacheEntryService);
export default CacheEntryService;
//# sourceMappingURL=CacheEntryService.js.map