import { produce } from 'immer';
import { applyMiddleware, createStore } from 'redux';
import createSagaMiddleware from 'redux-saga';
import useLazyRef from '../core/useLazyRef';
import { reducersSymbol, sagasSymbol } from '../types/app';
import { inReducer } from '../types/service';
import { clone, fromPayload, simpleMergeDeep } from '../utils/object';
import { defaultSettings, indexLocales } from './settings';
import useAppContext from './useAppContext';
import useRouter from './useRouter';
export const createReduxStore = (initialState = {}, middlewares = []) => {
    let sagaMiddleware = middlewares.find((middleware) => middleware.name === 'sagaMiddleware');
    if (!sagaMiddleware) {
        sagaMiddleware = createSagaMiddleware({
            onError: (e) => {
                throw e;
            },
        });
        middlewares.push(sagaMiddleware);
    }
    const store = createStore(() => {
        return initialState;
    }, applyMiddleware(...middlewares));
    const buildReducer = (store) => {
        return (state = store.getState(), action) => {
            const reducers = store[reducersSymbol][action.type];
            if (reducers) {
                const nextState = produce(state, (draftState) => {
                    Object.keys(reducers).forEach((k) => {
                        try {
                            reducers[k].bind[inReducer] = true;
                            reducers[k].bind.state = draftState;
                            reducers[k].func.apply(reducers[k].bind, fromPayload(action.payload));
                            if (action.resolve) {
                                setTimeout(() => {
                                    action.resolve(nextState);
                                }, 0);
                            }
                        }
                        catch (e) {
                            if (action.reject) {
                                action.reject(e);
                            }
                            throw e;
                        }
                        finally {
                            reducers[k].bind[inReducer] = false;
                        }
                    });
                });
                Object.keys(reducers).forEach((k) => {
                    reducers[k].bind.state = nextState;
                });
                return nextState;
            }
            return state;
        };
    };
    store[reducersSymbol] = {};
    store[sagasSymbol] = {};
    store.runSaga = (namespace, saga, name = 'root') => {
        store[sagasSymbol][`${namespace}.${name}`] = sagaMiddleware.run(saga);
    };
    store.cancelSaga = (namespace, name = 'root') => {
        const task = store[sagasSymbol][`${namespace}.${name}`];
        if (task) {
            task.cancel();
        }
    };
    store.injectReducers = (bind, namespace, reducers) => {
        Object.keys(reducers).forEach((actionType) => {
            store[reducersSymbol][actionType] = store[reducersSymbol][actionType] || {};
            store[reducersSymbol][actionType][`${namespace}.${actionType}`] = {
                func: reducers[actionType],
                bind,
            };
        });
        store.replaceReducer(buildReducer(store));
    };
    store.removeReducers = (namespace, reducers = {}) => {
        Object.keys(reducers).forEach((actionType) => {
            store[reducersSymbol][actionType] = store[reducersSymbol][actionType] || {};
            delete store[reducersSymbol][actionType][`${namespace}.${actionType}`];
        });
        store.replaceReducer(buildReducer(store));
    };
    return store;
};
export const formatSettings = (settings) => {
    let result = settings;
    if (Array.isArray(settings)) {
        result = Object.assign({}, settings[0]);
        for (let i = 1; i < settings.length; i++) {
            result = simpleMergeDeep(result, settings[i]);
        }
    }
    const cloneDefaultSettings = clone(defaultSettings);
    const formattedSettings = simpleMergeDeep(result, cloneDefaultSettings);
    return indexLocales(formattedSettings);
};
export function asResultCallback(callback, router, appContext) {
    if (!callback)
        return undefined;
    if (typeof callback === 'string') {
        return () => {
            router.push(callback);
        };
    }
    else {
        const originalCallback = callback;
        return (result) => {
            originalCallback(result, appContext);
        };
    }
}
function useResultCallback(callback) {
    const router = useRouter();
    const appContext = useAppContext();
    const callbackRef = useLazyRef(() => {
        return asResultCallback(callback, router, appContext);
    });
    return callbackRef.current;
}
export function useErrorCallback(callback, defaultCallback) {
    const resultCallback = useResultCallback(callback);
    const defaultResultCallback = useResultCallback(defaultCallback);
    return resultCallback || defaultResultCallback;
}
export function useSuccessCallback(callback, defaultCallback) {
    const resultCallback = useResultCallback(callback);
    const defaultResultCallback = useResultCallback(defaultCallback);
    return resultCallback || defaultResultCallback;
}
//# sourceMappingURL=utils.js.map