import { useCallback, useEffect } from 'react';
import useLocalService from '../app/useLocalService';
import { useErrorCallback, useSuccessCallback } from '../app/utils';
import useNotificationService from '../notification/useNotificationService';
import LoginService from './LoginService';
/**
 * The **useLogin** hooks instanciates a login service
 *
 * ```ts
 * const [error, loading, submit] = useLogin(idpName, options);
 * ```
 *
 * @param idpName **idpName** is used to retrieve the configuration identified by the key ***idp/:idpName*** in `src/settings.ts`.
 * @param options
 * @returns
 *
 * @group Hooks
 * @category Auth
 */
const useLogin = (idpName = 'default', options = {}) => {
    // create the local login service
    const [state, service] = useLocalService(LoginService, {
        loading: false,
    });
    // inject the global notificationService
    const notificationService = useNotificationService();
    // we send errors to the notification service
    const defaultOnError = useCallback((error) => {
        notificationService.send({
            topic: 'login-error',
            payload: error,
        });
    }, [notificationService]);
    const onError = useErrorCallback(options.onError) || defaultOnError;
    const onSuccess = useSuccessCallback(options.onSuccess);
    const callback = options.callback;
    // build the submit method in case of a form login
    const submit = useCallback((data) => {
        return service.formLogin(Object.assign({}, data), idpName, onError, onSuccess);
    }, [service, idpName, onError, onSuccess]);
    useEffect(() => {
        if (callback) {
            // call the external login callback saga
            service.externalLoginCallback(idpName, onError, onSuccess);
        }
        else {
            // call the login saga
            service.login(idpName, onError, onSuccess);
        }
    }, [service, idpName, onError, onSuccess, callback]);
    return [state.error, state.loading || false, submit];
};
// alias
export const useLoginService = useLogin;
export default useLogin;
//# sourceMappingURL=useLogin.js.map