import { __decorate, __metadata } from "tslib";
import { call, cancel, delay, fork, spawn } from 'redux-saga/effects';
import DefaultBasicError from '../core/BasicError';
import { HttpMethod } from '../types/fetch';
import { SagaEffect } from '../types/saga';
import { reducer, saga } from './annotations';
import DefaultService from './Service';
import { asyncHttp } from './xhr';
export default class FetchService extends DefaultService {
    pollTask;
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *delayLoading(delay_ms) {
        yield this.setLoading(delay_ms ? false : true, true);
        if (delay_ms) {
            yield delay(delay_ms);
            yield this.setLoading(true, true);
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *delete(url, options) {
        yield this.fetch(url, HttpMethod.Delete, undefined, options);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    fetchError(e) {
        this.state.error = e;
        this.setLoading(false, false);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    fetchSuccess(result) {
        this.state.error = undefined;
        this.state.result = result;
        this.setLoading(false, false);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *fetch(url, method, body, options = {}) {
        let loadingTask = null;
        try {
            loadingTask = yield fork([this, this.delayLoading], options.delayLoading);
            const fetcher = options.fetcher || asyncHttp;
            const result = yield fetcher(url, method, body, options);
            yield cancel(loadingTask);
            yield this.fetchSuccess(result); // to update the store and trigger a re-render.
            const onSuccess = options.onFetchSuccess;
            if (onSuccess) {
                yield call(onSuccess, result);
            }
        }
        catch (e) {
            if (process.env.NODE_ENV === 'development') {
                console.error('Fetch error', e);
            }
            if (loadingTask) {
                yield cancel(loadingTask);
            }
            yield this.fetchError(e);
            const onError = options.onFetchError;
            if (onError) {
                yield call(onError, DefaultBasicError.of(e));
            }
            else {
                throw e;
            }
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *get(url, options) {
        yield this.fetch(url, HttpMethod.Get, undefined, options);
    }
    *pollFetch(url, fixedRateInMs, options) {
        yield this.fetch(url, HttpMethod.Get, undefined, options);
        const interval = typeof fixedRateInMs === 'function' ? fixedRateInMs(this.state.result) : fixedRateInMs;
        if (interval > 0) {
            yield delay(interval);
            yield this.pollFetch(url, fixedRateInMs, options);
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *poll(url, fixedRateInMs, options) {
        this.pollTask = yield spawn([this, (this.pollFetch)], url, fixedRateInMs, options);
        return this.pollTask;
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *cancelPoll() {
        yield this.pollTask?.cancel();
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *patch(url, body, options) {
        yield this.fetch(url, HttpMethod.Patch, body, options);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *post(url, body, options) {
        yield this.fetch(url, HttpMethod.Post, body, options);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *put(url, body, options) {
        yield this.fetch(url, HttpMethod.Put, body, options);
    }
    setLoading(loading, fetching) {
        this.state.loading = loading;
        this.state.fetching = fetching;
    }
}
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "delete", null);
__decorate([
    reducer
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "fetchError", null);
__decorate([
    reducer
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "fetchSuccess", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, String, Object, Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "fetch", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", Object)
], FetchService.prototype, "get", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object, Object]),
    __metadata("design:returntype", Object)
], FetchService.prototype, "pollFetch", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object, Object]),
    __metadata("design:returntype", Object)
], FetchService.prototype, "poll", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "cancelPoll", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object, Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "patch", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object, Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "post", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object, Object]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "put", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Boolean, Boolean]),
    __metadata("design:returntype", void 0)
], FetchService.prototype, "setLoading", null);
//# sourceMappingURL=FetchService.js.map