import { produce } from 'immer';
import { apply, take } from 'redux-saga/effects';
import { combinedReducers, create, dispatch, inReducer, reducers, run, sagas, sagasFromReducers, stop, types, } from '../types/service';
import { ID } from '../types/symbol';
import { fromPayload, get, isGetterOrSetter, toPayload } from '../utils/object';
import { isFunction } from '../utils/type';
import DefaultBasicError from './BasicError';
const createReducer = Symbol('service.createReducer');
const createSaga = Symbol('service.createSaga');
export const handler = {
    get: function (target, prop, receiver) {
        const alias = target[types][prop];
        if (alias) {
            if (alias.type === 'reducer') {
                return function (...args) {
                    target[dispatch]({
                        type: alias.actionType,
                        payload: toPayload(args),
                    });
                };
            }
            else if (alias.type === 'saga') {
                return function (...args) {
                    return new Promise((resolve, reject) => {
                        target[dispatch]({
                            type: alias.actionType,
                            payload: toPayload(args),
                            resolve,
                            reject,
                        });
                    });
                };
            }
        }
        else {
            return Reflect.get(target, prop, receiver);
        }
    },
};
export default class DefaultService {
    [reducers];
    [types];
    [sagas];
    [dispatch];
    [combinedReducers] = (state) => state;
    [inReducer] = false;
    [sagasFromReducers];
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    state = null;
    [create](initialState) {
        this[reducers] = {};
        this[types] = {};
        this[sagas] = {};
        this[inReducer] = false;
        this[sagasFromReducers] = {};
        this.state = initialState;
        const createProperty = (property) => {
            if (isFunction(this[property]) && this[property].reducer) {
                this[createReducer](property, this[property]);
            }
            else if (isFunction(this[property]) && this[property].saga) {
                this[createSaga](property, this[property]);
            }
        };
        const properties = [];
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        let obj = this;
        while (obj) {
            const p = Object.getOwnPropertyNames(obj);
            // eslint-disable-next-line no-loop-func
            p.forEach((property) => {
                if (!isGetterOrSetter(obj, property) && !properties.includes(property)) {
                    createProperty(property);
                }
                properties.push(property);
            });
            obj = Object.getPrototypeOf(obj);
        }
        this[run]();
        if (isFunction(this.init)) {
            this[inReducer] = true;
            this.init();
            this[inReducer] = false;
        }
        // freeze state
        this.state = produce(this.state, (draftState) => draftState);
    }
    [createReducer](type, reducer) {
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        const self = this;
        const actionType = this.constructor[ID] ? `${this.constructor[ID]}.${type}` : type;
        if (reducer) {
            this[reducers][actionType] = reducer;
            this[types][type] = {
                type: 'reducer',
                actionType,
            };
        }
        this[type] = function (...args) {
            if (!self[inReducer]) {
                // call from a saga -> dispatch
                // return self[dispatch]({
                //   type: actionType,
                //   payload: toPayload(args),
                // });
                return new Promise((resolve, reject) => {
                    self[dispatch]({
                        type: actionType,
                        payload: toPayload(args),
                        resolve,
                        reject,
                    });
                });
            }
            return reducer.apply(self, args);
        };
    }
    /**
     * @param type name of the property
     * @param saga function with an additional value in the prototype : {
     *   saga: {
     *     effect: latest
     *     delay: 10
     *     defaultDelay: 20
     *   }
     * }
     */
    [createSaga](type, saga) {
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        const self = this;
        const actionType = this.constructor[ID] ? `${this.constructor[ID]}.${type}` : type;
        const effect = saga.saga.effect;
        let delay = saga.saga.delay;
        const defaultDelay = saga.saga.defaultDelay;
        if (typeof delay === 'string') {
            const sep = delay.indexOf('.');
            const obj = delay.substring(0, sep);
            if (sep === -1 || !['state', 'settings'].includes(obj)) {
                throw new DefaultBasicError('A string delay for a saga of type throttle or debounce must starts with state.xxx or settings.xxx');
            }
            if (obj === 'state') {
                delay = get(this.state, delay.substring(sep + 1), defaultDelay);
            }
            else if (this.context) {
                delay = get(this.context.settings, delay.substring(sep + 1), defaultDelay);
            }
        }
        const wrapper = function* wrapper(action) {
            try {
                const result = yield apply(self, saga, fromPayload(action.payload));
                if (action.resolve) {
                    action.resolve(result);
                }
                return result;
            }
            catch (error) {
                if (action.reject) {
                    action.reject(error);
                    return;
                }
                else {
                    throw error;
                }
            }
        };
        if (effect && effect !== take) {
            if (delay) {
                this[sagas][actionType] = function* () {
                    yield effect(delay, actionType, wrapper);
                };
            }
            else {
                this[sagas][actionType] = function* () {
                    yield effect(actionType, wrapper);
                };
            }
        }
        else {
            this[sagas][actionType] = function* () {
                while (true) {
                    const action = yield take(actionType);
                    yield wrapper(action);
                }
            };
        }
        this[types][type] = {
            type: 'saga',
            actionType,
        };
        this[type] = function* (...args) {
            yield saga.apply(self, args);
        };
        this[sagasFromReducers][type] = function (...args) {
            return new Promise((resolve, reject) => {
                if (self[dispatch]) {
                    self[dispatch]({
                        type: actionType,
                        payload: toPayload(args),
                        resolve,
                        reject,
                    });
                }
            });
        };
    }
    [run]() {
        this[combinedReducers] = (state, action) => {
            this[inReducer] = true;
            try {
                const nextState = produce(state, (draftState) => {
                    this.state = draftState;
                    if (this[reducers][action.type]) {
                        this[reducers][action.type].apply(this, fromPayload(action.payload));
                    }
                });
                this.state = nextState;
                return nextState;
            }
            finally {
                this[inReducer] = false;
            }
        };
    }
    [stop]() {
        this[combinedReducers] = (state) => state;
    }
    callSaga(sagaName, ...args) {
        return this[sagasFromReducers][sagaName](args);
    }
}
//# sourceMappingURL=Service.js.map