import { get, isDictionary, set } from '../utils/object';
import { urlBuilder } from '../utils/router';
import HTTPError from './HTTPError';
export const encodeFormData = (data) => {
    if (isDictionary(data)) {
        return Object.keys(data)
            .map((key) => encodeURIComponent(key) + '=' + encodeURIComponent(`${data[key]}`))
            .join('&');
    }
    return '';
};
export async function xhr(url, method, body, options = {}) {
    url = urlBuilder(url, options.params || {}, options.query || {});
    const fetchOptions = {
        method,
    };
    const headers = Object.assign({
        Accept: 'application/json',
    }, options.headers);
    if (['POST', 'PUT', 'PATCH'].includes(method)) {
        headers['Content-Type'] = headers['Content-Type'] || 'application/json';
        if (body) {
            if (headers['Content-Type'] === 'application/json') {
                fetchOptions.body = JSON.stringify(body);
            }
            else if (headers['Content-Type'] === 'application/x-www-form-urlencoded') {
                fetchOptions.body = encodeFormData(body);
            }
            else {
                throw Error(`Unsupported content-type ${headers['Content-Type']}`);
            }
        }
    }
    if (options.auth) {
        if (get(options, 'auth.token.access_token')) {
            headers.Authorization = `Bearer ${options.auth.token.access_token}`;
        }
        else if (get(options, 'auth.token')) {
            headers.Authorization = `Bearer ${options.auth.token}`;
        }
        else if (options.auth.basic) {
            headers.Authorization = `Basic ${btoa(options.auth.basic.user + ':' + options.auth.basic.password)}`;
        }
    }
    fetchOptions.headers = headers;
    // sanitize options
    const validOptions = [
        'method',
        'headers',
        'body',
        'mode',
        'credentials',
        'cache',
        'redirect',
        'referrer',
        'referrerPolicy',
        'integrity',
        'keepalive',
        'signal',
    ];
    Object.keys(options).forEach((k) => {
        if (validOptions.includes(k)) {
            set(fetchOptions, k, get(options, k));
        }
    });
    const result = await fetch(url, fetchOptions);
    return result;
}
export async function formatAsyncResponse(response) {
    const clone = response.clone();
    if (response.ok) {
        try {
            return await response.json();
        }
        catch (e) {
            return await clone.text();
        }
    }
    let result;
    try {
        result = await response.json();
    }
    catch (e) {
        throw new HTTPError(response.status, await clone.text());
    }
    throw new HTTPError(response.status, result.message, result);
}
export async function asyncHttp(url, method, body, options = {}) {
    return formatAsyncResponse(await xhr(url, method, body, options));
}
export async function asyncGet(url, options = {}) {
    return await asyncHttp(url, 'GET', undefined, options);
}
export async function asyncDelete(url, options = {}) {
    return await asyncHttp(url, 'DELETE', undefined, options);
}
export async function asyncPost(url, body, options = {}) {
    return await asyncHttp(url, 'POST', body, options);
}
export async function asyncPut(url, body, options = {}) {
    return await asyncHttp(url, 'PUT', body, options);
}
export async function asyncPatch(url, body, options = {}) {
    return await asyncHttp(url, 'PATCH', body, options);
}
//# sourceMappingURL=xhr.js.map