import { __decorate, __metadata } from "tslib";
import { all, call } from 'redux-saga/effects';
import DefaultGlobalService from '../app/GlobalService';
import { reducer, saga, service } from '../core/annotations';
import { asyncGet } from '../core/xhr';
import NotificationService from '../notification/NotificationService';
import { SagaEffect } from '../types/saga';
import { flattenTranslations } from '../utils/i18n';
import { append, get, set } from '../utils/object';
import { isLocaleSimple } from '../utils/settings';
import { isFunction } from '../utils/type';
let I18nService = class I18nService extends DefaultGlobalService {
    notificationService;
    modifiers;
    fetching;
    constructor(notificationService) {
        super();
        this.modifiers = {};
        this.fetching = {};
        this.notificationService = notificationService;
    }
    init() {
        const { settings } = this.context;
        this.modifiers = {
            locale: (value, locale) => {
                return value ? value.toLocaleString(locale) : value;
            },
        };
        if (settings.i18n && settings.i18n.modifiers) {
            Object.assign(this.modifiers, settings.i18n.modifiers);
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *changeLocale(locale) {
        const { router } = this.context;
        const location = router.location;
        if (isLocaleSimple(this.context.settings)) {
            yield this.setLocale(locale);
        }
        else {
            yield router.push(location, { locale });
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *fetchTranslations(locale, namespaces, options = {}) {
        // TODO description options
        const { settings } = this.context;
        const nsToFetch = namespaces.filter((ns) => !get(this.state, `i18n.fetching.${locale}.${ns}`, false));
        try {
            yield this.setFetchingTranslations(true, locale, nsToFetch);
            const results = yield all(nsToFetch.map((ns) => {
                return call(asyncGet, this.getLocaleUrl(locale, ns, settings), options);
            }));
            const translations = {};
            nsToFetch.forEach((ns, i) => (translations[ns] = results[i]));
            yield this.setTranslations(translations, locale); // to update the store and trigger a re-render.
            yield this.setFetchingTranslations(false, locale, nsToFetch);
        }
        catch (e) {
            yield this.setFetchingTranslations(false, locale, nsToFetch);
            const onError = options.onError;
            if (onError) {
                yield onError(e);
            }
            else {
                yield this.notificationService.error(e);
            }
        }
    }
    getLocaleUrl(locale, ns, settings) {
        if (isFunction(settings.i18n.url)) {
            return settings.i18n.url(locale, ns);
        }
        return `${settings.i18n.url}/${locale}/${ns}.json`;
    }
    setFetchingTranslations(fetching, locale, namespaces) {
        namespaces.forEach((ns) => set(this.state, `i18n.fetching.${locale}.${ns}`, fetching));
    }
    setLocale(locale) {
        set(this.state, 'i18n.locale', locale);
    }
    setTranslations(translations, locale) {
        set(this.state, `i18n.translations.${locale}`, Object.assign(get(this.state, `i18n.translations.${locale}`, {}), flattenTranslations(translations)));
        Object.keys(translations).forEach((ns) => append(this.state, `i18n.ns.${locale}`, ns));
    }
};
__decorate([
    saga(SagaEffect.Latest),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", void 0)
], I18nService.prototype, "changeLocale", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Array, Object]),
    __metadata("design:returntype", void 0)
], I18nService.prototype, "fetchTranslations", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Boolean, String, Array]),
    __metadata("design:returntype", void 0)
], I18nService.prototype, "setFetchingTranslations", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", void 0)
], I18nService.prototype, "setLocale", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, String]),
    __metadata("design:returntype", void 0)
], I18nService.prototype, "setTranslations", null);
I18nService = __decorate([
    service,
    __metadata("design:paramtypes", [NotificationService])
], I18nService);
export default I18nService;
//# sourceMappingURL=I18nService.js.map