import React, { useCallback, useEffect, useMemo } from 'react';
import useAppContext from '../app/useAppContext';
import useGlobalProp from '../app/useGlobalProp';
import DefaultBasicError from '../core/BasicError';
import { stringifyJsx } from '../utils/jsx';
import { get } from '../utils/object';
import useI18nService from './useI18nService';
import useLocale from './useLocale';
import { buildJsx } from './utils';
const useTranslation = (namespaces = []) => {
    const locale = useLocale();
    const appContextTranslations = get(useAppContext(), 'i18n.translations');
    const reduxTranslations = useGlobalProp(`i18n.translations.${locale}`);
    const translations = useMemo(() => {
        return Object.assign({}, appContextTranslations, reduxTranslations);
    }, [appContextTranslations, reduxTranslations]);
    const appContextNs = get(useAppContext(), 'i18n.ns');
    const reduxNs = useGlobalProp(`i18n.ns.${locale}`);
    const nsLoaded = useMemo(() => {
        return Array.from(new Set([...(appContextNs || []), ...(reduxNs || [])]));
    }, [appContextNs, reduxNs]);
    const i18nService = useI18nService();
    const nsRequired = useMemo(() => {
        let nsRequired = namespaces || [];
        if (typeof namespaces === 'string') {
            nsRequired = [namespaces];
        }
        if (!nsRequired.includes('common')) {
            nsRequired.push('common');
        }
        return nsRequired;
    }, [namespaces]);
    const nsNotLoaded = useMemo(() => {
        return nsRequired.filter((ns) => !nsLoaded.includes(ns));
    }, [nsRequired, nsLoaded]);
    const nsFetching = useGlobalProp(`i18n.fetching.${locale}`, {});
    const fetching = !!Object.keys(nsFetching).find((ns) => nsFetching[ns] && nsNotLoaded.includes(ns));
    const t = useCallback((content, alias, count = 1) => {
        if (fetching)
            return null;
        let key = alias;
        let aliasNs;
        if (alias) {
            // check if alias contains a namespace
            const aliasTokens = alias.split(':');
            if (aliasTokens.length > 2) {
                throw new DefaultBasicError(`The i18n alias '${alias}' is invalid. It can't contain the character ':'`);
            }
            else if (aliasTokens.length === 2) {
                aliasNs = aliasTokens[0];
                if (!nsRequired.includes(aliasNs)) {
                    throw new DefaultBasicError(`The i18n alias '${alias}' is invalid. The namespace ${aliasNs} is not declared as a namespace by useTranslation`);
                }
                key = aliasTokens[1];
            }
        }
        const ctx = {};
        let jsx = null;
        const candidateKeys = [];
        if (typeof content !== 'string') {
            jsx = content;
            const [jsxKey] = stringifyJsx(content, ctx);
            if (!alias) {
                key = jsxKey;
            }
        }
        else {
            if (!alias) {
                key = content;
            }
        }
        if (count >= 1) {
            const prefix = count > 1 ? 'plural' : 'singular';
            // candidateKeys.push(`${prefix}::${key}`);
            nsRequired.forEach((ns) => {
                if (!aliasNs || aliasNs === ns) {
                    candidateKeys.push(`${ns}:${prefix}::${key}`);
                }
            });
        }
        // candidateKeys.push(key as string);
        nsRequired.forEach((ns) => {
            if (!aliasNs || aliasNs === ns) {
                candidateKeys.push(`${ns}:${key}`);
            }
        });
        for (const candidateKey of candidateKeys) {
            if (translations[candidateKey] !== undefined) {
                const result = buildJsx(translations[candidateKey], ctx, jsx, i18nService, locale);
                return result;
            }
        }
        return null;
    }, [fetching, translations, nsRequired, i18nService, locale]);
    const T = useCallback(({ alias, count, children }) => {
        return t(React.createElement(React.Fragment, null, children), alias, count);
    }, [t]);
    useEffect(() => {
        if (nsNotLoaded.length > 0 && !fetching) {
            i18nService.fetchTranslations(locale, nsNotLoaded);
        }
    }, [nsNotLoaded, i18nService, locale, fetching]);
    return [T, t, locale, fetching || nsNotLoaded.length > 0];
};
export default useTranslation;
//# sourceMappingURL=useTranslation.js.map