import { __decorate, __metadata } from "tslib";
import { delay } from 'redux-saga/effects';
import DefaultGlobalService from '../app/GlobalService';
import { reducer, saga, service } from '../core/annotations';
import { SagaEffect } from '../types/saga';
import { append, get, isNull, set } from '../utils/object';
import { NotificationLevel } from './typings';
import { isNotificationContent } from './utils';
let nextId = 1;
let NotificationService = class NotificationService extends DefaultGlobalService {
    init() {
        this.context.router.listen(this.onRouteChange);
    }
    add(notification) {
        const max = get(this.context.settings, `notification.${notification.topic}.max`, get(this.context.settings, `notification.default.max`, 0));
        append(this.state, `notifications.${notification.topic}`, notification);
        if (max > 0 && this.state.notifications[notification.topic].length > max) {
            this.state.notifications[notification.topic].shift();
        }
    }
    clearTopic(topic) {
        set(this.state, `notifications.${topic}`, []);
    }
    clearAll() {
        this.state.notifications = {};
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *debug(payload) {
        yield this.send(this.formatLevelNotification(NotificationLevel.Debug, payload));
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *error(payload) {
        yield this.send(this.formatLevelNotification(NotificationLevel.Error, payload));
    }
    formatNotification(notificationContent, settings, notificationService) {
        if (!isNotificationContent(notificationContent)) {
            notificationContent = {
                payload: notificationContent,
            };
        }
        const persist = get(settings, `notification.${notificationContent.topic}.persist`, get(settings, `notification.default.persist`, true));
        if (notificationContent.ttl === undefined) {
            notificationContent.ttl = get(settings, `notification.${notificationContent.topic}.ttl`, get(settings, `notification.default.ttl`));
        }
        const timestamp = notificationContent.timestamp === undefined ? new Date().getTime() : notificationContent.timestamp;
        const id = notificationContent.id === undefined ? `oneki-id-${++nextId}` : notificationContent.id;
        const topic = notificationContent.topic === undefined ? 'default' : notificationContent.topic;
        const expires = notificationContent.ttl ? timestamp + notificationContent.ttl : null;
        const notification = {
            id,
            topic,
            timestamp,
            persist: notificationContent.persist === undefined ? persist : notificationContent.persist,
            ttl: notificationContent.ttl,
            remove: () => {
                notificationService.remove(id);
            },
            expires,
            payload: notificationContent.payload,
        };
        return notification;
    }
    formatLevelNotification(level, payload) {
        return {
            topic: level,
            payload,
        };
    }
    getNotification(notificationId) {
        let notification = undefined;
        Object.keys(this.state.notifications || {}).forEach((topic) => {
            for (let i = 0; i < this.state.notifications[topic].length; i++) {
                if (this.state.notifications[topic][i].id === notificationId) {
                    notification = this.state.notifications[topic][i];
                    break;
                }
            }
        });
        return notification;
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *info(payload) {
        yield this.send(this.formatLevelNotification(NotificationLevel.Info, payload));
    }
    onRouteChange() {
        Object.keys(this.state.notifications || {}).forEach((topic) => {
            for (let i = this.state.notifications[topic].length - 1; i >= 0; i--) {
                if (!this.state.notifications[topic][i].persist) {
                    this.state.notifications[topic].splice(i, 1);
                }
            }
        });
    }
    remove(notificationId) {
        Object.keys(this.state.notifications || {}).forEach((topic) => {
            for (let i = 0; i < this.state.notifications[topic].length; i++) {
                if (this.state.notifications[topic][i].id === notificationId) {
                    this.state.notifications[topic].splice(i, 1);
                    break;
                }
            }
        });
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *send(notificationContent) {
        const { store, settings } = this.context;
        if (isNotificationContent(notificationContent) && !isNull(notificationContent.id)) {
            // check if this notification is already present
            const topic = notificationContent.topic || 'default';
            const notifications = get(store.getState(), `notifications.${topic}`, []);
            if (notifications.find((n) => n.id === notificationContent.id)) {
                return;
            }
        }
        const notification = this.formatNotification(notificationContent, settings, this);
        yield this.add(notification);
        yield this._delayRemove(notification);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *success(payload) {
        yield this.send(this.formatLevelNotification(NotificationLevel.Success, payload));
    }
    /**
     * This method is called whenever the auto-removal of the notification should be suspended
     * @param notificationId
     */
    touch(notificationId, permanent = false) {
        const notification = this.getNotification(notificationId);
        if (notification && notification.ttl && notification.expires) {
            notification.ttl = permanent ? undefined : notification.expires - new Date().getTime();
            notification.expires = null;
        }
    }
    /**
     * This method is called whenever the auto-removal of the notification should be reactivated
     * @param notificationId
     */
    *untouch(notificationId) {
        const notification = this.getNotification(notificationId);
        if (notification && notification.ttl) {
            this._setExpires(notification.id, new Date().getTime() + notification.ttl);
            yield this._delayRemove(notification);
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    *warning(payload) {
        yield this.send(this.formatLevelNotification(NotificationLevel.Warning, payload));
    }
    _setExpires(notificationId, expires) {
        const notification = this.getNotification(notificationId);
        if (notification) {
            notification.expires = expires;
        }
    }
    *_delayRemove(notification) {
        if (notification.ttl) {
            yield delay(notification.ttl);
            const refreshedNotification = this.getNotification(notification.id);
            if (refreshedNotification && refreshedNotification.expires) {
                // if ttl > 0 and notifcation.expires is null, it means that the removal is suspended
                yield this.remove(refreshedNotification.id);
            }
        }
    }
};
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "add", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "clearTopic", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "clearAll", null);
__decorate([
    saga(SagaEffect.Every)
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "debug", null);
__decorate([
    saga(SagaEffect.Every)
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "error", null);
__decorate([
    saga(SagaEffect.Every)
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "info", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "onRouteChange", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "remove", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "send", null);
__decorate([
    saga(SagaEffect.Every)
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "success", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "touch", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "untouch", null);
__decorate([
    saga(SagaEffect.Every)
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    ,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "warning", null);
__decorate([
    reducer,
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "_setExpires", null);
__decorate([
    saga(SagaEffect.Every),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", void 0)
], NotificationService.prototype, "_delayRemove", null);
NotificationService = __decorate([
    service
], NotificationService);
export default NotificationService;
//# sourceMappingURL=NotificationService.js.map